from yattag import Doc
from usdm_excel.cross_ref import cross_references
from usdm_excel.logger import package_logger
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance, ScheduledInstance
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit

import traceback

class ExportAsTimeline():

  FULL = "full"
  BODY = "body"

  def __init__(self, study):
    self.study = study

  def export(self, level=FULL):
    try:
      doc = Doc()
      study_design = self.study.versions[0].studyDesigns[0]
      if level == self.BODY:
        self._body(doc, study_design)
      else:
        self._full(doc, study_design)
      return doc.getvalue()
    except Exception as e:
      package_logger.error(f"Exception [{e}] raised generating HTML page {level}.")
      package_logger.debug(f"{traceback.format_exc()}")

  def _full(self, doc, study_design):
    doc.asis('<!DOCTYPE html>')
    with doc.tag('html'):
      with doc.tag('head'):
        pass
      with doc.tag('body'):
        self._body(doc, study_design)
  
  def _body(self, doc, study_design):
    for timeline in study_design.studyScheduleTimelines:
      timings = []
      with doc.tag(f'h1'):
        doc.asis(f'{timeline.name}')
      with doc.tag('pre', klass='mermaid'):
        doc.asis('\ngraph LR\n')
        doc.asis(f'{timeline.id}([{timeline.entryCondition}])\n')
        instance = cross_references.get_by_id(ScheduledActivityInstance, timeline.scheduleTimelineEntryId)
        if instance.instanceType == 'ACTIVITY': 
          doc.asis(f'{instance.id}(A)\n')
        else:
          doc.asis(f'{instance.id}{{{{D}}}}\n')
        doc.asis(f'{timeline.id} -->|first| {instance.id}\n')
        for timing in instance.scheduledInstanceTimings:
          #print(f"APPEND: {timing}")
          timings.append(timing)
        prev_instance = instance
        instance = cross_references.get_by_id(ScheduledActivityInstance, instance.defaultConditionId)
        while instance:
          #print(f"INST: {instance}")
          if instance.instanceType == 'ACTIVITY': 
            doc.asis(f'{instance.id}(A)\n')
          else:
            doc.asis(f'{instance.id}{{{{D}}}}\n')
            for condition in instance.conditionAssignments:
              doc.asis(f'{instance.id} -->|{condition[0]}| {condition[1]}\n') 
          doc.asis(f'{prev_instance.id} -->|default| {instance.id}\n')      
          for timing in instance.scheduledInstanceTimings:
            #print(f"APPEND: {timing}")
            timings.append(timing)
          prev_instance = instance
          instance = self._get_cross_reference(prev_instance.defaultConditionId)
          #print(f"NEXT: {instance}")
        exit = cross_references.get_by_id(ScheduleTimelineExit, prev_instance.scheduleTimelineExitId)
        doc.asis(f'{exit.id}([Exit])\n')
        doc.asis(f'{prev_instance.id} -->|exit| {exit.id}\n')      
        for timing in timings:
          #print(f"TIMING: {timing}")
          doc.asis(f'{timing.id}(({timing.label}\n{timing.type.decode}\n{timing.timingValue}\n{timing.timingWindowLower}..{timing.timingWindowUpper}))\n')            
          doc.asis(f'{timing.id} -->|from| {timing.relativeFromScheduledInstanceId}\n')      
          doc.asis(f'{timing.id} -->|to| {timing.relativeToScheduledInstanceId}\n')      
    with doc.tag('script', type='module'):
      doc.asis("import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';\n")   
      doc.asis("mermaid.initialize({ startOnLoad: true });\n")   

  def _get_cross_reference(self, id):
    for klass in [ScheduledActivityInstance, ScheduledDecisionInstance]:
      instance = cross_references.get_by_id(klass, id)
      if instance:
        return instance 
    return None




import requests
import os
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.cdisc_ct_library import cdisc_ct_library
from usdm_excel.ncit import NCIt
from usdm_excel.id_manager import id_manager
from usdm_excel.alias import Alias
from usdm_model.biomedical_concept import BiomedicalConcept
from usdm_model.biomedical_concept_property import BiomedicalConceptProperty
from usdm_model.response_code import ResponseCode
from usdm_model.alias_code import AliasCode
from usdm_excel.logger import package_logger

class CDISCBiomedicalConcepts():

  API_ROOT = 'https://api.library.cdisc.org/api/cosmos/v2'    
  
  def __init__(self):
    self.api_key = os.getenv('CDISC_API_KEY')
    self.headers =  {"Content-Type":"application/json", "api-key": self.api_key}
    self.package_metadata = self._get_package_metadata()
    self.package_items = self._get_package_items()
    self._bc_responses = {}

  def exists(self, name):
    name_uc = name.upper() # Avoid case mismatches
    if name_uc in self.package_items:
      return self.package_items[name_uc]
    else:
      return None

  def catalogue(self) -> list:
    return list(self.package_items.keys())
  
  def synonyms(self, name) -> list:
    metadata = self.exists(name)
    if not metadata:
      return []
    else:
      bc = self.usdm(name)
      return bc.bcSynonyms

  def to_cdisc_json(self, name) -> dict:
    metadata = self.exists(name)
    if not metadata:
      return {}
    else:
      return self._get_bc(metadata['href'])

  def to_usdm_json(self, name) -> dict:
    metadata = self.exists(name)
    if not metadata:
      return {}
    else:
      bc = self.usdm(name)
      return bc.to_json()

  def usdm(self, name) -> BiomedicalConcept:
    metadata = self.exists(name)
    if not metadata:
      return None
    else:
      response = self._get_bc(metadata['href'])
      bc = self._bc_as_usdm(response)
      if 'dataElementConcepts' in response:
        for item in response['dataElementConcepts']:
          codes = []
          if 'exampleSet' in item:
            for example in item['exampleSet']:
              term = cdisc_ct_library.preferred_term(example)
              if term != None:
                codes.append(CDISCCT().code(term['conceptId'], term['preferredTerm']))
          bc.bcProperties.append(self._bc_property_as_usdm(item, codes))
      return bc

  def _get_package_metadata(self) -> dict:
    api_url = self._url('/mdr/bc/packages')
    package_logger.info("CDISC BC Library: %s" % api_url)
    try:
      raw = requests.get(api_url, headers=self.headers)
      response = raw.json()
      packages = response['_links']['packages']
      return packages
    except:
      package_logger.info("CDISC BC Library FAILED: %s" % api_url)
      return {}

  def _get_package_items(self) -> dict:
    results = {}
    try:
      for package in self.package_metadata:
        api_url = self._url(package['href']) 
        package_logger.info("CDISC BC Library: %s" % api_url)
        raw = requests.get(api_url, headers=self.headers)
        response = raw.json()
        for item in response['_links']['biomedicalConcepts']:
          results[item['title'].upper()] = item
      return results
    except:
      results = {}

  def _url(self, relative_url) -> str:
    return "%s%s" % (self.__class__.API_ROOT, relative_url)

  def _bc_as_usdm(self, api_bc) -> BiomedicalConcept:
    code = NCIt().code(api_bc['conceptId'], api_bc['shortName'])
    synonyms = api_bc['synonyms'] if 'synonyms' in api_bc else []
    return BiomedicalConcept(
      id=id_manager.build_id(BiomedicalConcept),
      name=api_bc['shortName'],
      label=api_bc['shortName'],
      bcSynonyms=synonyms,
      bcReference=api_bc['_links']['self']['href'],
      bcProperties=[],
      code=Alias().code(code, [])
    )

  def _bc_property_as_usdm(self, property, codes) -> BiomedicalConceptProperty:
    concept_code = NCIt().code(property['conceptId'], property['shortName'])
    concept_aliases = []
    responses = []
    for code in codes:
      responses.append(ResponseCode(id=id_manager.build_id(ResponseCode), responseCodeEnabled=True, code=code))
    return BiomedicalConceptProperty(
      id=id_manager.build_id(BiomedicalConceptProperty),
      name=property['shortName'],
      label=property['shortName'],
      bcPropertyRequired=True,
      bcPropertyEnabled=True,
      bcPropertyDatatype=property['dataType'],
      bcPropertyResponseCodes=responses,
      code=Alias().code(concept_code, concept_aliases)
    )

  def _get_bc(self, url):
    if url in self._bc_responses:
      return self._bc_responses[url]
    else:
      api_url = self._url(url)
      raw = requests.get(api_url, headers=self.headers)
      result = raw.json()
      self._bc_responses[url] = result
      return result

cdisc_bc_library = CDISCBiomedicalConcepts()


from typing import List, Union
from pydantic import Field, constr
from .api_base_model import ApiBaseModel
from .study_protocol_document import StudyProtocolDocument
from .study_version import StudyVersion
from uuid import UUID

class Study(ApiBaseModel):
  id: Union[UUID, None] = None
  name: str = constr(min_length=1)
  description: Union[str, None] = None
  label: Union[str, None] = None
  versions: List[StudyVersion] = []
  documentedBy: Union[StudyProtocolDocument, None] = None



from usdm_model.organization import Organization
from usdm_model.address import Address
from usdm_model.study_identifier import StudyIdentifier
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.iso_3166 import ISO3166
from usdm_excel.cross_ref import cross_references
import pandas as pd
import traceback

class StudyIdentifiersSheet(BaseSheet):

  def __init__(self, file_path: str):
    try:
      super().__init__(file_path=file_path, sheet_name='studyIdentifiers')
      self.identifiers = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")
      
  def process_sheet(self):
    self.identifiers = []
    for index, row in self.sheet.iterrows():
      org_type = self.read_cdisc_klass_attribute_cell_by_name('Organization', 'organizationType', index, ['organisationType', 'type'])     
      org_id_scheme = self.read_cell_by_name(index, 'organisationIdentifierScheme')
      org_identifier = self.read_cell_by_name(index, 'organisationIdentifier')
      org_name = self.read_cell_by_name(index, ['organisationName', 'name'])
      org_label = self.read_cell_by_name(index, 'label', default="")
      org_address = self._build_address(index)
      try:
        organisation = Organization(
          id=id_manager.build_id(Organization),
          organizationIdentifierScheme=org_id_scheme, 
          organizationIdentifier=org_identifier,
          name=org_name,
          label=org_label,
          type=org_type,
          organizationLegalAddress=org_address
        )
      except Exception as e:
        self._general_error(f"Failed to create Organization object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")
      else:
        cross_references.add(organisation.id, organisation)   
        try:
          item = StudyIdentifier(
            id=id_manager.build_id(StudyIdentifier),
            studyIdentifier=self.read_cell_by_name(index, 'studyIdentifier'), 
            studyIdentifierScope=organisation
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyIdentifier object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
        else:
          self.identifiers.append(item)
          cross_references.add(item.studyIdentifier, item)         
  def _build_address(self, row_index):
    field_name = 'organisationAddress'
    raw_address = self.read_cell_by_name(row_index, field_name)
    # TODO The '|' separator is preserved for legacy reasons but should be removed in the future
    if '|' in raw_address:
      sep = '|'
      parts = raw_address.split(sep)
    else:
      sep = ','
      parts = self._state_split(raw_address)
    if len(parts) == 6:
      # TODO Put something in each part if empty. Temp fix
      for index, part in enumerate(parts):
        parts[index] = '-' if part == '' else part.strip()
      result = self._to_address(
          id_manager.build_id(Address),
          line=parts[0], 
          district=parts[1], 
          city=parts[2], 
          state=parts[3], 
          postal_code=parts[4], 
          country=ISO3166().code(parts[5])
        )
      return result
    else:
      col_index = self.sheet.columns.get_loc(field_name)
      self._error(row_index, col_index, f"Address does not contain the required fields (line, district, city, state, postal code and country code) using '{sep}' separator characters, only {len(parts)} found")
      return None

  def _to_address(self, id, line, city, district, state, postal_code, country):
    text = "%s, %s, %s, %s, %s, %s" % (line, city, district, state, postal_code, country.decode)
    text = text.replace(' ,', '')
    try:
      result = Address(id=id, text=text, line=line, city=city, district=district, state=state, postalCode=postal_code, country=country)
    except Exception as e:
      self._general_error(f"Failed to create Address object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      result = None
    else:
      cross_references.add(result.id, result)   
    return result


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_v2_sheet.conditons import Conditons
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.encounter import Encounter
from usdm_model.study_epoch import StudyEpoch
from usdm_model.activity import Activity
import traceback

class ScheduledInstance():
  
  def __init__(self, parent, col_index):
    self.parent = parent
    self.item = None
    self.col_index = col_index
    epoch_id = None
    encounter_id = None
    name = self.parent.read_cell(SoAColumnRows.NAME_ROW, col_index)
    self.name = name
    description = self.parent.read_cell(SoAColumnRows.DESCRIPTION_ROW, col_index)
    label = self.parent.read_cell(SoAColumnRows.LABEL_ROW, col_index)
    epoch_name = self.parent.read_cell(SoAColumnRows.EPOCH_ROW, col_index)
    encounter_name = self.parent.read_cell(SoAColumnRows.ENCOUNTER_ROW, col_index)
    #print(f"ENCOUNTER: {encounter_name}")
    type = self.parent.read_cell(SoAColumnRows.TYPE_ROW, col_index)
    self.default_name = self.parent.read_cell(SoAColumnRows.DEFAULT_ROW, col_index)
    self.conditions = Conditons(self.parent.read_cell(SoAColumnRows.CONDITIONS_ROW, col_index))
    if encounter_name:
      encounter = cross_references.get(Encounter, encounter_name)
      if encounter:
        encounter_id = encounter.id
        #print(f"ENCOUNTER2: {encounter_id}")
      else:
        self.parent._general_warning(f"Failed to find encounter with name '{encounter_name}'")
    if epoch_name:
      epoch = cross_references.get(StudyEpoch, epoch_name)
      if epoch:
        epoch_id = epoch.id
      else:
        self.parent._general_warning(f"Failed to find epoch with name '{epoch_name}'")
    try:
      if type.upper() == "ACTIVITY":
        self.item = ScheduledActivityInstance(
          id=id_manager.build_id(ScheduledActivityInstance),
#          name=name,
#          description=description,
#          label=label,
          instanceType='ACTIVITY',
          scheduleTimelineExitId=None,
          scheduledActivityInstanceEncounterId=encounter_id,
          scheduledInstanceTimings=[],
          scheduledInstanceTimelineId=None,
          defaultConditionId=None,
          epochId=epoch_id,
          activityIds=self._add_activities()
        )
        #print(f"ENCOUNTER3: {self.item}")
        cross_references.add(self.item.id, self.item)
      elif type.upper() == "DECISION":
        self.item = ScheduledDecisionInstance(
          id=id_manager.build_id(ScheduledDecisionInstance),
          instanceType='DECISION',
          scheduleTimelineExitId=None,
          #scheduledInstanceEncounterId=None,
          scheduledInstanceTimings=[],
          scheduledInstanceTimelineId=None,
          defaultConditionId=None,
          conditionAssignments=[]
        )
        cross_references.add(self.item.id, self.item)
      else:
        self.parent._general_warning(f"Unrecognized ScheduledInstance type: '{type}'")
    except Exception as e:
      self.parent._general_error(f"Exception [{e}] raised reading sheet")
      self.parent._traceback(f"{traceback.format_exc()}")

  def _add_activities(self):
    activities = []
    row = 0
    column = self.parent.sheet.iloc[:, self.col_index]
    for cell in column:
      if row >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity_name = self.parent.read_cell(row, SoAColumnRows.CHILD_ACTIVITY_COL)
        if str(cell).upper() == "X":
          activity = cross_references.get(Activity, activity_name)
          if activity:
            activities.append(activity.id)
          else:
            self.parent._general_warning(f"Unable to find activity '{activity_name}' when adding to schedule instance")
      row += 1
    return activities




from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_excel.study_soa_v2_sheet.activities import Activities
from usdm_excel.study_soa_v2_sheet.scheduled_instances import ScheduledInstances
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit

import traceback

class StudySoAV2Sheet(BaseSheet):

  NAME_ROW = 0
  DESCRIPTION_ROW = 1
  CONDITION_ROW = 2
  PARAMS_DATA_COL = 1

  def __init__(self, file_path, sheet_name, main=False, require={}):
    try:
      super().__init__(file_path=file_path, sheet_name=sheet_name, header=None, require=require)
      self.name = ""
      self.description = ""
      self.condition = ""
      self.timeline = None
      self.main_timeline = main
      self.encounters = []
      self.activities = []
      self.timelines = []
      self.biomedical_concept_surrogates = []
      self.biomedical_concepts = []
      self._process_sheet()
      self._raw_instances = ScheduledInstances(self)
      self._raw_activities = Activities(self)

      # @TODO Move block to Activities class
      for item in self._raw_activities.items:
        activity = item.usdm_activity
        self.activities.append(activity)
        self.biomedical_concept_surrogates += item.usdm_biomedical_concept_surrogates
        self.biomedical_concepts += item.usdm_biomedical_concepts
      self.double_link(self.activities, 'previousActivityId', 'nextActivityId')
      
      self.timeline = self._add_timeline(self.name, self.description, self.condition, self._raw_instances.instances, self._raw_instances.exits)

    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet")
      self._traceback(f"{traceback.format_exc()}")

  def set_timing_references(self, timings):
    for timing in timings:
      instance = self._raw_instances.match(timing.relativeFromScheduledInstanceId)
      if instance:
        item = instance.item
        timing.relativeFromScheduledInstanceId = item.id
        item.scheduledInstanceTimings.append(timing)
      # else:
      #   self._general_error(f"Unable to find timing 'from' reference with name {timing.relativeFromScheduledInstanceId}")
      instance = self._raw_instances.match(timing.relativeToScheduledInstanceId)
      if instance:
        item = instance.item
        timing.relativeToScheduledInstanceId = item.id
      # else:
      #   self._general_error(f"Unable to find timing 'to' reference with name {timing.relativeToScheduledInstanceId}")
    for instance in self._raw_instances.items:
      item = instance.item
      if isinstance(item, ScheduledActivityInstance):
        found = False
        for timing in timings:
          ids = [timing.relativeFromScheduledInstanceId, timing.relativeToScheduledInstanceId]
          if item.id in ids:
            found = True
            break
        if not found:
          self._general_warning(f"Unable to find timing reference for instance with name {instance.name}")

  def _process_sheet(self):
    for rindex in range(self.NAME_ROW, self.CONDITION_ROW + 1):
      if rindex == self.NAME_ROW:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.DESCRIPTION_ROW:
        self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.CONDITION_ROW:
        self.condition = self.read_cell(rindex, self.PARAMS_DATA_COL)
      else:
        pass

  def _add_timeline(self, name, description, condition, instances, exit):
    try:
      timeline = ScheduleTimeline(
        id=id_manager.build_id(ScheduleTimeline),
        mainTimeline=self.main_timeline,
        name=name,
        description=description,
        label=name,
        entryCondition=condition,
        scheduleTimelineEntryId=instances[0].id,
        scheduleTimelineExits=exit,
        scheduleTimelineInstances=instances
      )
      cross_references.add(timeline.name, timeline)
      return timeline
    except Exception as e:
      self._general_error(f"Failed to create ScheduleTimeline object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      return None







from typing import List, Union
from .api_base_model import ApiBaseModelWithId
from .study_identifier import *
from .study_protocol_document_version import *
from .alias_code import *
from .code import Code as genericCode
from .study_design import *
from .governance_date import GovernanceDate
from .study_amendment import StudyAmendment

class StudyVersion(ApiBaseModelWithId):
  studyTitle: str
  studyVersion: str
  studyRationale: str
  studyAcronym: str
  type: Union[genericCode, None] = None
  studyPhase: Union[AliasCode, None] = None
  documentVersionId: Union[str, None] = None
  dateValues: List[GovernanceDate] = []
  amendments: List[StudyAmendment] = []
  businessTherapeuticAreas: List[Code] = []
  studyIdentifiers: List[StudyIdentifier] = []
  studyDesigns: List[StudyDesign] = []



from typing import List
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .biomedical_concept_property import BiomedicalConceptProperty

class BiomedicalConcept(ApiBaseModelWithIdNameAndLabel):
  bcSynonyms: List[str] = []
  bcReference: str
  bcProperties: List[BiomedicalConceptProperty] = []
  code: AliasCode



from typing import Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class StudyEpoch(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  previousStudyEpochId: Union[str, None] = None
  nextStudyEpochId: Union[str, None] = None



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.eligibility_criteria import EligibilityCriteria
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary

import traceback
import re

class StudyDesignEligibilityCriteriaSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignEligibilityCriteria', optional=True)
      self.items = []
      if self.success:
        for index, row in self.sheet.iterrows():
          category = self.read_cdisc_klass_attribute_cell_by_name('EligibilityCriteria', 'category', index, 'category')
          identifier = self.read_cell_by_name(index, 'identifier')
          name = self.read_cell_by_name(index, 'name')
          description = self.read_cell_by_name(index, 'description')
          label = self.read_cell_by_name(index, 'label')
          text = self.read_cell_by_name(index, 'text')
          dictionary_name = self.read_cell_by_name(index, 'dictionary')
          self._validate_references(index, 'text', text, dictionary_name)
          criteria = self._criteria(name, description, label, text, category, identifier, dictionary_name)
          if criteria:
            self.items.append(criteria)
        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _criteria(self, name, description, label, text, category, identifier, dictionary_name):
    try:
      dictionary_id = self._get_dictionary_id(dictionary_name)
      # dictionary = cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      # if dictionary:
      #   dictionary_id = dictionary.id
      # else:
      #   self._general_error(f"Unable to find dictionary with name '{dictionary_name}'")
      #   dictionary_id = None
      item = EligibilityCriteria(
        id=id_manager.build_id(EligibilityCriteria),
        instanceType='ELIGIBILITY_CRITERIA', 
        name=name,
        description=description,
        label=label,
        text=text,
        category=category,
        identifier=identifier,
        dictionaryId=dictionary_id
      )
    except Exception as e:
      self._general_error(f"Failed to create EligibilityCriteria object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      return None
    else:
      cross_references.add(item.id, item)
      return item

  def _validate_references(self, row, column_name, text, dictionary_name):
    if dictionary_name:
      column = self.column_present(column_name)
      dictionary = cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if not dictionary:
        self._warning(row, column, f"Dictionary '{dictionary_name}' not found")
        return
      tags = re.findall(r'\[([^]]*)\]',text)
      for tag in tags:
        if not tag in dictionary.parameterMap:
          self._warning(row, column, f"Failed to find '{tag}' in dictionary '{dictionary_name}'")
  
  def _get_dictionary_id(self, dictionary_name):
    if dictionary_name:
      dictionary = cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if dictionary:
        return dictionary.id
      else:
        self._general_error(f"Unable to find dictionary with name '{dictionary_name}'")
    return None



from typing import List
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent

class StudyProtocolDocumentVersion(ApiBaseModelWithId):
  briefTitle: str
  officialTitle: str
  publicTitle: str
  scientificTitle: str
  protocolVersion: str
  protocolStatus: Code
  dateValues: List[GovernanceDate] = []
  contents: List[NarrativeContent] = []
  childrenIds: List[str] = []



from typing import List, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .study_protocol_document_version import StudyProtocolDocumentVersion

class StudyProtocolDocument(ApiBaseModelWithIdNameLabelAndDesc):
  versions: List[StudyProtocolDocumentVersion] = []



from typing import Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class Timing(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  timingValue: str
  timingRelativeToFrom: Code
  relativeFromScheduledInstanceId: Union[str, None] = None
  relativeToScheduledInstanceId: Union[str, None] = None
  timingWindowLower: Union[str, None] = None
  timingWindowUpper: Union[str, None] = None
  timingWindow: Union[str, None] = None



from typing import List
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .response_code import ResponseCode

class BiomedicalConceptProperty(ApiBaseModelWithIdNameAndLabel):
  bcPropertyRequired: bool
  bcPropertyEnabled: bool
  bcPropertyDatatype: str
  bcPropertyResponseCodes: List[ResponseCode] = []
  code: AliasCode



from .activity import *
from .address import *
from .alias_code import *
from .analysis_population import *
from .api_base_model import *
from .biomedical_concept_category import *
from .biomedical_concept_property import *
from .biomedical_concept_surrogate import *
from .biomedical_concept import *
from .code import *
from .eligibility_criteria import *
from .encounter import *
from .endpoint import *
from .subject_enrollment import *
from .estimand import *
from .geographic_scope import *
from .governance_date import *
from .indication import *
from .intercurrent_event import *
from .investigational_intervention import *
from .narrative_content import *
from .objective import *
from .organization import *
from .procedure import *
from .response_code import *
from .schedule_timeline_exit import *
from .schedule_timeline import *
from .scheduled_instance import *
from .study_amendment import *
from .study_amendment_reason import *
from .study_arm import *
from .study_cell import *
from .study_design_population import *
from .study_design import *
from .study_element import *
from .study_epoch import *
from .study_identifier import *
from .study_protocol_document_version import *
from .study_protocol_document import *
from .study_version import *
from .study import *
from .syntax_template import *
from .syntax_template_dictionary import *
from .timing import *
from .transition_rule import *

__all__ = [
  'Activity',
  'Address',
  'AliasCode',
  'AnalysisPopulation',
  'BiomedicalConceptCategory',
  'BiomedicalConceptProperty',
  'BiomedicalConceptSurrogate',
  'BiomedicalConcept',
  'Code',
  'EligibilityCriteria',
  'Encounter',
  'Endpoint',
  'Enrollment',
  'Estimand',
  'GeographicScope',
  'GovernanceDate',
  'Indication',
  'IntercurrentEvent',
  'InvestigationalIntervention',
  'NarrativeContent',
  'Objective',
  'Organization',
  'Procedure',
  'ResponseCode',
  'ScheduleTimelineExit',
  'ScheduleTimeline',
  'ScheduledInstance',
  'StudyAmendment',
  'StudyAmendmentReason',
  'StudyArm',
  'StudyCell',
  'StudyDesignPopulation',
  'StudyDesign',
  'StudyElement',
  'StudyEpoch',
  'StudyIdentifier',
  'StudyProtocolDocumentVersion',
  'StudyProtocolDocument',
  'Study',
  'StudyVersion',
  'SyntaxTemplate',
  'SyntaxTemplateDictionary',
  'Timing',
  'TransitionRule',
]


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
import traceback
from usdm_model.study_design_population import StudyDesignPopulation

class StudyDesignPopulationSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignPopulations')
      self.populations = []
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, 'name', default=f"POP {index + 1}") # 'name' added, preserve backward compatibility so defaulted
        description = self.read_description_by_name(index, ['description', 'populationDescription']) # Allow multiple names for column
        label = self.read_cell_by_name(index, 'label', default="")
        number = self.read_cell_by_name(index, "plannedNumberOfParticipants")
        min = self.read_cell_by_name(index, "plannedMinimumAgeOfParticipants")
        max = self.read_cell_by_name(index, "plannedMaximumAgeOfParticipants")
        codes = self._build_codes(row, index)
        try:
          pop = StudyDesignPopulation(id=id_manager.build_id(StudyDesignPopulation),
            name=name,
            description=description,
            label=label,
            plannedNumberOfParticipants=number,
            plannedMinimumAgeOfParticipants=min,
            plannedMaximumAgeOfParticipants=max,
            plannedSexOfParticipants=codes
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyDesignPopulation object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
        else:
          cross_references.add(name, pop)
          self.populations.append(pop)
        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      #print(f"{traceback.format_exc()}")
      self._traceback(f"{traceback.format_exc()}")

  def _build_codes(self, row, index):
    code = self.read_cdisc_klass_attribute_cell_by_name('StudyDesignPopulation', "plannedSexOfParticipants", index, "plannedSexOfParticipants")
    return [code] if code else []
      

        


import yaml
import csv
from src.usdm_excel import USDMExcel
from uuid import UUID

SAVE_ALL = True

def to_int(value):
  try:
    return int(value)
  except:
    return None

def read_error_csv(file):
  reader = csv.DictReader(file)
  items = list(reader)
  for item in items:
    item['row'] = to_int(item['row'])
    item['column'] = to_int(item['column'])
  return items

def run_test(filename, mocker, save=False):
  fake_uuids = (UUID(f'00000000-0000-4000-8000-{i:012}', version=4) for i in range(10000))
  mocker.patch("usdm_excel.export_as_neo4j_dict.uuid4", side_effect=fake_uuids)
  excel = USDMExcel(f"tests/integration_test_files/{filename}.xlsx")
  result = excel.to_neo4j_dict()

  # Useful if you want to see the results.
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}_neo4j_dict.yaml", 'w') as f:
      f.write(yaml.dump(result))
  
  with open(f"tests/integration_test_files/{filename}_neo4j_dict.yaml", 'r') as f:
    expected = yaml.safe_load(f) 
  assert result == expected

def test_simple_1(mocker):
  run_test('simple_1', mocker)

def test_full_1(mocker):
  run_test('full_1', mocker)



from src.usdm_excel.cross_ref import *

class CRTest():

  def __init__(self, id, name):
    self.id = id
    self.name = name

def test_create():
  object = CrossRef()
  assert len(object.references.keys()) == 0
  assert object.references == {}

def test_add():
  item = CRTest(id="1234", name="name")
  cross_references.references = {}
  cross_references.add("name", item)
  assert len(cross_references.references.keys()) == 1
  assert cross_references.references["CRTest.name"] == item
  assert len(cross_references.identifiers.keys()) == 1
  assert cross_references.identifiers["CRTest.1234"] == item

def test_get():
  item = CRTest(id="1234", name="name")
  cross_references.references = {}
  cross_references.references["CRTest.name"] = item
  assert cross_references.get(CRTest, "name") == item

def test_get_by_id():
  item = CRTest(id="1234", name="name")
  cross_references.identifiers = {}
  cross_references.identifiers["CRTest.1234"] = item
  assert cross_references.get_by_id(CRTest, "1234") == item
  assert cross_references.get_by_id("CRTest", "1234") == item

def test_clear():
  item = CRTest(id="1234", name="name")
  cross_references.references = {}
  cross_references.identifiers = {}
  cross_references.references["CRTest.name"] = item
  cross_references.identifiers["CRTest.1234"] = item
  assert len(cross_references.references.keys()) == 1
  assert len(cross_references.identifiers.keys()) == 1
  cross_references.clear()
  assert len(cross_references.references.keys()) == 0
  assert len(cross_references.identifiers.keys()) == 0



from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.code import Code

class CodeBase():

  def _build(self, code, system, version, decode):
    instance = Code(id=id_manager.build_id(Code), code=code, codeSystem=system, codeSystemVersion=version, decode=decode)
    cross_references.add(instance.id, instance)
    return instance
 


import re
from usdm_excel.iso_8601_duration import ISO8601Duration

class WindowType():

  def __init__(self, timing_info):
    self.upper = None
    self.lower = None
    self.errors = []
    self.label = timing_info.strip()
    if timing_info:
      match = re.match(r"(?P<lower>[+|-]*\d+)\s*\.\.\s*(?P<upper>[+|-]*\d+) \s*(?P<units>.+)", self.label)
      if match is not None:
        timing_parts = match.groupdict()
        self.lower = self._set_encoded(f"{timing_parts['lower']}{timing_parts['units']}")
        self.upper = self._set_encoded(f"{timing_parts['upper']}{timing_parts['units']}")     
      else:
        self.errors.append(f"Could not decode the window value, not all required parts detected in '{timing_info}'")
    else:
      pass # Empty, this is OK

  def _set_encoded(self, duration):
    the_duration = duration.strip()
    original_duration = the_duration
    for char in ['+', '-']:
      if char in the_duration:
        the_duration = the_duration.replace(char, "")
        #self.warnings.append(f"Ignoring '{char}' in {original_duration}")
    duration_parts = re.findall(r"[^\W\d_]+|\d+", the_duration)
    if len(duration_parts) == 2:
      try:
        value = ISO8601Duration().encode(duration_parts[0].strip(), duration_parts[1].strip())
        return value
      except Exception as e:
        self.errors.append(f"Could not decode the duration value '{the_duration}'")
        return ""
    else:
      self.errors.append(f"Could not decode the duration value, no value and units found in '{the_duration}'")
      return ""



from typing import Union
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .code import Code
from .address import Address

class Organization(ApiBaseModelWithIdNameAndLabel):
  type: Code
  organizationIdentifierScheme: str
  organizationIdentifier: str
  organizationLegalAddress: Union[Address, None] = None
